/* 
 *    Bounding-Box collision detection demo
 *    by Neil Chinnery
 */



#include <stdlib.h>
#include <stdio.h>
#include <sys/farptr.h>
#include <sys/segments.h>
#include <sys/movedata.h>
#include "allegro.h"
#include "sprites.h" /* Data file Index - contains some sprites and a pallete */
#define MAX_SPRITES 10

DATAFILE *data;
BITMAP *s;

struct spr_bb{ /* Sprite Bounding-boxes */
 int bb_height; /* Bounding-box height of sprite */
 int bb_width; /* Bounding-box width of sprite */
} spr_bb[MAX_SPRITES];

int x1=20, y1=20; /* X and Y co-ords of Sprite 1 */
int x2=150, y2=150; /* X and Y co-ors of Sprite 2 */
int sp1=1, sp2=2; /* Sprite 1 & 2 positions in datafile */



void init_sp_bb() /* initialise data in BB to be zeros */
{
 int l;

 for (l=0; l<(MAX_SPRITES+1); l++)
 {
  spr_bb[l].bb_height=0;
  spr_bb[l].bb_width=0;
 }
}

void mk_spr_bb(BITMAP *s3, int spr_no)
{
 int bb_height=0; /* Bounding box height of sprite */
 int bb_width=0; /* Bounding box width of sprite */
 int x1, y1; /* Used to span through pixels within the sprite */
 int p; /* Holds return value from getpixel() */

         
 for (y1=0; y1<128; y1++)      /* Now go through each pixel of the sprite */
 {
  for (x1=0; x1<128; x1++)       /* in turn */
  {
   p=getpixel(s3,x1,y1);         /* Find out what colour the pixel is */
   if ((p!=0) && (p!=-1))        /* if the colour is non-zero */
    {
     if (y1>bb_height) bb_height=y1;  /* if > than current bounding box width, increase bounding box width */
     if (x1>bb_width) bb_width=x1;    /* if > than current bounding box heigth, increase bounding box height */
    }
  }
 }
 spr_bb[spr_no].bb_height=bb_height; /* Set the final bounding box height and width */
 spr_bb[spr_no].bb_width=bb_width;
}

int check_collision(int spr1, int spr1x, int spr1y, int spr2, int spr2x, int spr2y)
 {
 if ((spr1x>spr2x+spr_bb[spr2].bb_width) || (spr2x > spr1x+spr_bb[spr1].bb_width) ||
     (spr1y>spr2y+spr_bb[spr2].bb_height) || (spr2y> spr1y+spr_bb[spr1].bb_height))
  {
     return 0;
  }
  else
  {
   return 1;
  }
}

void mve_spr()
{
 if (key[KEY_UP]) y1--;
 if (key[KEY_DOWN]) y1++;
 if (key[KEY_RIGHT]) x1++;
 if (key[KEY_LEFT]) x1--;
 if (key[KEY_1])
 {
  sp1++;
  clear_keybuf();
 }
 if (key[KEY_2])
 {
  sp2++;
  clear_keybuf();
 }
 if (y1<0) y1=0;
 if (y1>460) y1=460;
 if (x1<0) x1=0;
 if (x1>620) x1=620;
 if (sp1>7) sp1=1;
 if (sp2>7) sp2=1;

}

void drw_spr()
{
 char msg[80];
 clear_to_color(s,0);
 draw_sprite (s, data[sp1].dat, x1, y1);
 draw_sprite (s, data[sp2].dat, x2, y2);
 sprintf(msg, "Sprite (1)=%d", sp1);
 textout(s,font,msg, 0,250,255);
 sprintf(msg, "Bounding Box Height=%d", spr_bb[sp1].bb_height);
 textout(s,font,msg, 0,270,255);
 sprintf(msg, "Bounding Box Width=%d", spr_bb[sp1].bb_width);
 textout(s,font,msg, 0,290,255);

 sprintf(msg, "Sprite (2)=%d", sp2);
 textout(s,font,msg, 300,250,255);
 sprintf(msg, "Bounding Box Height=%d", spr_bb[sp2].bb_height);
 textout(s,font,msg, 300,270,255);
 sprintf(msg, "Bounding Box Width=%d", spr_bb[sp2].bb_width);
 textout(s,font,msg, 300,290,255);

 /* here's where we test for collision! */

 if (check_collision(sp1,x1,y1,sp2,x2,y2))
    textout(s,font, "Bounding Boxes Interset - Collision Detected!!",150,320,255);
 vsync();
 blit(s,screen,0,0,0,0,640,480);
}


void main()
{
 int c;
 char msg[80];


   init_sp_bb(); /* Initialise bounding-box structure */

   /* Initialise Allegro Units */
   allegro_init();
   install_keyboard(); 
   install_timer();

   data=load_datafile("sprites.dat"); /* Load the datafile */
   
   clear_keybuf(); /* Clear the keyboard buffer */
   
   /* Create bitmap to allow double-buffering */
   s=create_bitmap(640,480);

   clear_to_color(s,0);



   /* Create Bounding Boxes for the seven sprites in the datafile */

  for (c=1; c<8; c++) mk_spr_bb(data[c].dat, c);

  /* Set the Graphics mode */

   set_gfx_mode(GFX_AUTODETECT, 640, 480, 0, 0);
   set_pallete(data[PAL1].dat);
   do
   {
   mve_spr();  /* Move the sprites */
   drw_spr(); /*Display the sprites */
   }while(!key[KEY_ESC]);
   clear_keybuf();
   exit(0);
}
