/* Copyright 1997  Jonathan Griffiths. See "readme.jon" for detailed terms. */
#include <stdlib.h>
#include <jlib.h>

/* Allows resizing sprite files and shows how to resize sprites on the fly. */

void usage(void);
void update_zoom(sprite_system *sys,sprite_system *save,int zoomx,int zoomy);

int main(int argc,char **argv)
{
	sprite_system *sys,*save;
	buffer_rec *offscreen;
	int i,zoomx,zoomy,start=0;
	UBYTE *pal=NULL;

	JLIB_DEBUG_OFF;

	if ((argc != 3) && (argc != 4))
		usage();

	save = sprite_init(1,255);

	if (sprite_load(argv[1],save) != SUCCESS) {
		puts("ERROR: Couldn't load input file.");
		usage();
	}

	sys = sprite_init(1,SPR_NUM_LOADED(save));

	if (sprite_load(argv[1],sys) != SUCCESS) {
		puts("ERROR: Couldn't load input file.");
		usage();
	}

	sprite_free(save);

	save = sprite_init(1,SPR_NUM_LOADED(sys));

	if (sprite_load(argv[1],save) != SUCCESS) {
		puts("ERROR: Couldn't load input file.");
		usage();
	}

	if (argc == 4) {
		pal = pal_load(argv[3]);
		if (pal == NULL) {
			puts("ERROR: Couldn't load palette file.");
			usage();
		}
	}

	screen_set_video_mode();

	if (pal != NULL)
		screen_block_set_pal(pal);

	if (!KEYBOARD_AVAILABLE)
		jlib_exit("No Keyboard Support.");

	zoomx = sys->sprite_data[0]->width;
	zoomy = sys->sprite_data[0]->height;

	offscreen = buff_init(SCREEN_WIDTH,SCREEN_HEIGHT);
	screen_blit_fs_buffer(offscreen);

	kb_init();

	while (!kb_keydown(KEY_ESC)) {
		buff_clear(offscreen);

		for (i = start;i<SPR_NUM_LOADED(sys);i++)
			buff_stamp_sprite(sys,i,offscreen,(i-start)*SPR_MAX_X,1);

		screen_blit_fs_buffer(offscreen);

		if (kb_keydown(KEY_DOWN))
			if(zoomy<SPR_MAX_Y-2){
				zoomy++;
				update_zoom(sys,save,zoomx,zoomy);
			}

		if (kb_keydown(KEY_UP))
			if(zoomy>1){
				zoomy--;
				update_zoom(sys,save,zoomx,zoomy);
			}

		if (kb_keydown(KEY_RIGHT))
			if(zoomx<SPR_MAX_X-2){
				zoomx++;
				update_zoom(sys,save,zoomx,zoomy);
			}

		if (kb_keydown(KEY_LEFT))
			if(zoomx>1){
				zoomx--;
				update_zoom(sys,save,zoomx,zoomy);
			}

		if(kb_keydown(KEY_Q))
			if(start>1)
				start--;

		if(kb_keydown(KEY_A))
			if(start<(SPR_NUM_LOADED(sys)-1))
				start++;
	}
	
	kb_closedown();

	sprite_save(argv[2],sys);

	screen_restore_video_mode();

	return 1;
}


/* display usage information */
void usage(void)
{
	puts("spr_size.exe: Resize the sprites in a file");
	puts("Usage: spr_size.exe <sprite_file> <output_file> [<pal_file>]");
	puts("Where <pal_file> is optional.");
	puts("Use the arrow keys to resize your sprites.");
	puts("Q and A keys move through the sprites.");
	puts("ESC quits and saves output to <output_file>.");

	exit(0);
}


/* rebuilds the sprite system dynamically as a different size */
void update_zoom(sprite_system *sys,sprite_system *save,int zoomx,int zoomy)
{
	buffer_rec *one,*two;
	int i;

	one = buff_init(SPR_MAX_X,SPR_MAX_Y);
	two = buff_init(SPR_MAX_X,SPR_MAX_Y);

	for (i=0;i<SPR_NUM_LOADED(save);i++) {
		buff_clear(one);
		buff_clear(two);
		buff_stamp_sprite(save,i,one,0,0);
		buff_scale_buff_to(two,0,0,zoomx,zoomy,one,0,0,save->sprite_data[i]->width,save->sprite_data[i]->height);
		sprite_build_from_buff(sys,i,two,0,0,zoomx,zoomy);
	}

	buff_free(one);
	buff_free(two);
}
