/* Copyright 1995-96 Jon Griffiths.  See the file "jlib.doc" for details. */
#ifndef LIST_H
#define LIST_H

#include <stdlib.h>
#include <jlib.h>

/* List data structures -------------------------------------------------- */
typedef struct list_node{
	void *data;
        struct list_node *prev;
        struct list_node *next;
}list_node;

typedef struct{
        list_node *head;
        list_node *tail;
        list_node *current;
}list;



/* Set up the list ------------------------------------------------------- */
list *list_init();

/* Destroy the list ------------------------------------------------------ */
list *list_destroy(list *list);

/* Add an item to the start of the list ---------------------------------- */
void list_add_to_end(list *list,void *list_item);

/* Add an item to the end of the list ------------------------------------- */
void list_add_to_start(list *list,void *list_item);

/* Add an item after the current item in the list ------------------------ */
void list_add_after_current(list *list,void *list_item);

/* add an item before the current item in the list ----------------------- */
void list_add_before_current(list *list,void *list_item);

/* Remove the current item in the list ----------------------------------- */
void list_remove_current(list *list);

/* Remove the head of the list ------------------------------------------- */
void list_remove_head(list *list);

/* remove the tail of the list ------------------------------------------- */
void list_remove_tail(list *list);


/* Return the data pointer of a node ------------------------------------- */
#define list_get_data_pointer(node) ((node)->data)

/* Move the current pointer to the next item if one exists --------------- */
#define list_move_current_forward(l) if((l)->current->next!=NULL)(l)->current=(l)->current->next

/* Move the current pointer to the previous item if one exists ----------- */
#define list_move_current_backward(l) if((l)->current->prev!=NULL)(l)->current=(l)->current->prev


/* Get the current pointer ----------------------------------------------- */
#define list_get_current(l)          (l)->current

/* Set the current pointer ----------------------------------------------- */
#define list_set_current(l,ptr)      (l)->current=(ptr)

/* Is the current item the last in the list? ----------------------------- */
#define list_current_at_end(l)       ((l)->current==(l)->tail)

/* Is the current item the first in the list? ---------------------------- */
#define list_current_at_start(l)     ((l)->current==(l)->head)

/* Is the current item NULL? --------------------------------------------- */
#define list_current_is_null(l)      ((l)->current==NULL)

/* Set the current item to the first item -------------------------------- */
#define list_set_current_to_start(l) ((l)->current=(l)->head)

/* Set the current item to the last item --------------------------------- */
#define list_set_current_to_end(l)   ((l)->current=(l)->tail)

/* Set the current item to NULL ------------------------------------------ */
#define list_set_current_to_null(l)  ((l)->current=NULL)

#endif

