/*
* SPIN.C -*- Nissim Chudnoff -*- Sprite by John P. Griffiths
*****************************************************************************
* Compare Speed Of (in order):                                              *
* screen_blit_fs_buffer()                                                   *
* screen_blit_buff_to()                                                     *
* screen_blit_buff_toNC()                                                   *
* screen_blit_buff_to() (Dirty Rectangle)                                   *
* screen_blit_buff_toNC() (Dirty Rectangle)                                 *
****************************************************************************/

/*
RESULTS:
MACHINE: i386 sx/16, 4MB Ram, Compressed with DJP 1.04, ISA VGA card 256K RAM
*****************************************************************************
* screen_blit_fs_buffer()                     | 11.22 Seconds               *
* screen_blit_buff_to()                       | 10.75 Seconds               *
* screen_blit_buff_toNC()                     | 10.70 Seconds               *
* screen_blit_buff_to() (Dirty Rectangle)     |  2.15 Seconds               *
* screen_blit_buff_toNC() (Dirty Rectangle)   |  2.09 Seconds               *
****************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <jlib.h>

#define NUM_ITERATIONS 100

int cnt;        /* Counter For Iterations */
char times = 1;     /* Counter For blit Type */
sprite_system *demo_sprites;
buffer_rec *buff;

int main()
{
UBYTE *pal;
int animation[6] = {0,1,2,3,4,5};

/* init stuff*/
buff = buff_init(SCREEN_WIDTH,SCREEN_HEIGHT);
demo_sprites = sprite_init(1,6);
pal = screen_load_pal("spin.pal");
 
if( sprite_load("spin.spr",demo_sprites) != SUCCESS )
{
        puts("error: Couldn't load sprites");
        exit(-1);
}
 
if( (buff == NULL) || (demo_sprites == NULL) || (pal == NULL) )
{
        puts("initialisation error.");
        exit(-1);
}
 
screen_set_video_mode();

screen_block_set_pal(pal);

/* set up sprite movements in whole mode steps */
sprite_set_xy(demo_sprites,0,50,50);
sprite_set_an_frame(demo_sprites,0,0);
sprite_set_move_info(demo_sprites,0,0,0,0);
sprite_set_anim_info(demo_sprites,0,1,6,animation);
sprite_turn_on(demo_sprites,0);

fgetc(stdin);

/* this is the main sprite drawing loop */
do {
        buff_save_all_sprites(demo_sprites,buff);
        buff_draw_all_sprites(demo_sprites,buff);
        do_blit();
        buff_rest_all_sprites(demo_sprites,buff);
        sprite_update_all_anim_and_move(demo_sprites);
        cnt++;
}while( cnt != -1 );

screen_restore_video_mode();
 
}

int do_blit() /* Procedure For Blit Handling */
{
if( times == 1 )
{
        screen_blit_fs_buffer(buff);
        if( cnt == NUM_ITERATIONS )
        {
                fgetc(stdin);
                times++;
                cnt = 0;
        }
        return;
}
else if( times == 2 )
{
        screen_blit_buff_to(0,0,buff,0,0,SCREEN_MAX_X,SCREEN_MAX_Y);
        if( cnt == NUM_ITERATIONS )
        {
                fgetc(stdin);
                times++;
                cnt = 0;
        }
        return;
}
else if( times == 3 )
{
        screen_blit_buff_toNC(0,0,buff,0,0,SCREEN_MAX_X,SCREEN_MAX_Y);
        if( cnt == NUM_ITERATIONS )
        {
                fgetc(stdin);
                times++;
                cnt = 0;
        }
        return;
}
else if( times == 4 )
{
        screen_blit_buff_to(50,50,buff,50,50,114,106);
        if( cnt == NUM_ITERATIONS )
        {
                fgetc(stdin);
                times++;
                cnt = 0;
        }
        return;
}
else if( times == 5 )
{
        screen_blit_buff_to(50,50,buff,50,50,114,106);
        if( cnt == NUM_ITERATIONS )
        {
                cnt = -2;       /* Remeber the cnt++; at end of loop! */
                fgetc(stdin);
        }
        return;
}
}
