
#ifndef _FIXED_H_
#define _FIXED_H_

#include <limits.h>

typedef long fixed;

// fixed point conversions
#define INT_TO_FIXED(i) (((fixed) (i)) << 8)
#define FIXED_TO_INT(f) ((f) >> 8)
#define FIXED_TO_FLOAT(f) (((double) (f)) / 256)
#define FLOAT_TO_FIXED(f) ((fixed) ((f) * 256.0))

// functions
#define FIXED_ABS(f) ((f) < 0 ? -(f) : (f))
#define FIXED_TRUNC(f) ((f) & 0xffff0000)
#define FIXED_SIGN(f) ((unsigned long) (f) >> 31)
#define FIXED_PRODUCT_SIGN(f, g) ((unsigned long) ((f) ^ (g)) >> 31)
#define FIXED_HALF(f) ((f) / 2)
#define FIXED_DOUBLE(f) ((f) << 1)

// perform integer scaling of a fixed point number
#define FIXED_SCALE(f, i) ((f) * ((fixed) (i)))


// fixed point constants
#define FIXED_ZERO     (INT_TO_FIXED(0))
#define FIXED_ONE      (INT_TO_FIXED(1))
#define FIXED_ONE_HALF (FIXED_HALF(FIXED_ONE))
#define FIXED_PI       (FLOAT_TO_FIXED(3.14159265))
#define FIXED_2PI      (FLOAT_TO_FIXED(6.28318531))
#define FIXED_HALF_PI  (FLOAT_TO_FIXED(1.57079633))
#define FIXED_MIN      LONG_MIN
#define FIXED_MAX      LONG_MAX

// we need this for kludges to avoid fixed point division overflow
#define FIXED_EPSILON  ((fixed) 0x100)


// Alternate fixed point formats for use when extra fractional precision is needed.
#define TO_FIX_2_30(f)   ((f) << 14)
#define FROM_FIX_2_30(f) ((f) >> 14) 
#define TO_FIX_8_24(f)   ((f) << 8)
#define FROM_FIX_8_24(f) ((f) >> 8) 



/*********************** Intel fixed point functions **********************/

static inline fixed fixmul(fixed r1, fixed r2)
{
	fixed result;

   __asm__ ("imull %2\n\t"
            "shrd  $8, %%edx, %%eax\n\t"
	         :"=a" (result):"a" (r1), "d" (r2):"eax", "edx");

   return result;
}


static inline fixed fixmul2_30(fixed r1, fixed r2)
{
	fixed result;

   __asm__ ("imull %2\n\t"
	         "shrd  $30, %%edx, %%eax\n\t"
	         :"=a" (result):"a" (r1), "d" (r2):"eax", "edx");

   return result;
}


static inline fixed fixdiv(fixed dividend, fixed divisor)
{
	fixed result;

   // no checking for overflow is performed
   __asm__("movl %%edx, %%eax\n\t"
           "sar  $8, %%edx\n\t"
           "shl  $8, %%eax\n\t"
	        "idivl %%ecx, %%eax\n\t"
	        :"=a" (result):"d" (dividend), "c" (divisor):"eax", "ecx", "edx");

   return result;
}


#endif 
