/***************************************************************************

    Well, I was working on an Animation format myself, but decided to
    make life easier.  Here is a FLI/FLC player for DjGpp (I used the
    v2 beta-3 of the compiler, it may or maynot work on 1.??m? of it)
    using the GRX (again the v2 beta-3) library.

    If you can use it, kewl.  If not, oh well.  It was ported to DjGpp
    from a FLX (modified FLI) library for XLib v6.??.  This code was
    originally part of his FLI->FLX program, but has been easily worked
    into the GrContext setup.

    This will set the image at the center of the screen, but I'd make
    sure you have a big enough screen image to view it on.  (Obviously)
    You must set the mode prior to calling play_fli(), or who knows what
    will happen.

    Hope you like it, email me with whatever.

    And if anyone can come up with some decent speed detection and
    delaying code, send me some mail!

    John Tarbox
    -----------
    School  - tarboxj@merwan.sunyerie.edu
    BBS     - battle.axe@pegasuz.com

*/
#include <stdio.h>
#include <mem.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <grx/grx.h>   /* I set this up in my include directory */


#define USHORT unsigned short
#define SHORT short

#define FLIH_MAGIC_1 0xaf11           /* File header Magic */
#define FLIH_MAGIC_2 0xaf12           /* File header Magic (Pro) */
#define FLIF_MAGIC   0xf1fa           /* Frame Magic */

typedef struct fli_head	{
    long size           __attribute__ ((packed));
    USHORT type         __attribute__ ((packed));  /* = FLIH_MAGIC */
    USHORT frame_count  __attribute__ ((packed));
    USHORT width        __attribute__ ((packed));
    USHORT height       __attribute__ ((packed));
    USHORT bits_a_pixel __attribute__ ((packed));
    SHORT flags         __attribute__ ((packed));
    SHORT speed         __attribute__ ((packed));
    long next_head      __attribute__ ((packed));
   long frames_in_table __attribute__ ((packed));
    short file          __attribute__ ((packed)); /* used by players.  Contains zeros on disk. */
    long frame1_off     __attribute__ ((packed)); /* used by players.  Contains zeros on disk. */
    long strokes        __attribute__ ((packed)); /* how many paint strokes etc. made. */
    long session        __attribute__ ((packed)); /* stokes since file's been loaded. */
    char reserved[88]   __attribute__ ((packed)); /* all zeroes on disk */
} Fli_head;

/* bit defines for flags field */
#define FLI_FINISHED 1	/* finished writing fli */
#define FLI_LOOPED	2	/* fli has a loop frame */

typedef struct fli_frame {
    long size       __attribute__ ((packed));
    USHORT type     __attribute__ ((packed)); /* = 0xf1fa FLIF_MAGIC */
    SHORT chunks    __attribute__ ((packed));
    char pad[8]     __attribute__ ((packed));
} Fli_frame;


typedef struct fli_chunk {
    long size       __attribute__ ((packed));
    SHORT type      __attribute__ ((packed));
} Fli_chunk;


static Fli_head flihead;
static Fli_frame fliframehead;
static int flifile;


int setupfliread(char *fnam) {
	flifile = open(fnam,O_RDWR|O_BINARY,S_IWRITE|S_IREAD);
	if (flifile<=0) return 1;

	read(flifile,&flihead,sizeof(flihead));
    if (flihead.type!=FLIH_MAGIC_1)
        if (flihead.type!=FLIH_MAGIC_2)
            return 2;
	return 0;
	}

int readframe(unsigned char *thisframe)
	{
    unsigned char *flibuf,*p;
	Fli_chunk *chunk;
	unsigned short offs,len,skip,packets;
	char size;
    int i,j,k,x;

	read(flifile,&fliframehead,sizeof(fliframehead));
    if (fliframehead.type==0x00a1) {  /* Check for FLC's useless frame */
        long tmp;
        tmp = tell(flifile);
        lseek(flifile,tmp+fliframehead.size,SEEK_SET);
        return -1;
    }

	if (fliframehead.size==sizeof(fliframehead)) return 0;
    flibuf = (unsigned char *)malloc(fliframehead.size-sizeof(fliframehead));
	if (flibuf==NULL) return 2;
	read(flifile,flibuf,(unsigned short)fliframehead.size-sizeof(fliframehead));

	/* Now for the decoding */
	p = flibuf;
	for (i=0; i<fliframehead.chunks; i++) {
		chunk = (Fli_chunk *)p;
		p += sizeof(Fli_chunk);
		switch (chunk->type) {
            case 4:    /* FLI_256_COLOR */
				packets = *((short *)p); p+=2;
				for (j=0; j<packets; j++) {
                    int r,g,b,z;
					offs = *(p++);
					len  = *(p++); if (len==0) len = 256;
                    for(z=0; z<len; z++) {
                        r = *(p++);
                        g = *(p++);
                        b = *(p++);
                        GrSetColor(offs+z,r,g,b);
                    }
                }
				break;
            case 7:    /* FLI_DELTA */
                /* len     =  number of compressed lines
                 * packets =  number of lines to skip (if<0)
                 * packets =  number of packets to decode (if>0)
                 */
                len  = *((short *)p); p+= 2;
                for (j=0; j<flihead.height;) {
                    short packs;
                    packs = *(p++);
                    if (packs < 0)
                        j += -packs;
                    else if (packs > 0) {
                        x = 0;
                        for (k=0; k<packs; k++) {
                            skip = *(p++);
                            if (skip) x += skip;
                            size = *((char *)p); p++;
                            if (size>0) {
                                memcpy(thisframe+j*flihead.width+x,p,size*2);
                                p += (int)size*2;
                                x += (int)size*2;
                            }
                            else if (size<0) {
                                int z;
                                for (z=0; z<-size;z++) {
                                    memset(thisframe+j*flihead.width+x+(z*2),*p,1);
                                    memset(thisframe+j*flihead.width+x+((z*2)+1),*(p+1),1);
                                }
                                x -= (int)size*2;
                                p+=2;
                            }
                        }
                        j++;
                    }
                }
				break;
			case 11:	/* FLI_COLOR */
				packets = *((short *)p); p+=2;
				for (j=0; j<packets; j++) {
                    int r,g,b,z;
					offs = *(p++);
					len  = *(p++); if (len==0) len = 256;
                    for(z=0; z<len; z++) {
                        r = *(p++);
                        g = *(p++);
                        b = *(p++);
                        GrSetColor(offs+z,r*4,g*4,b*4);
                    }
                }
				break;
			case 12:	/* FLI_LC */
				offs = *((short *)p); p+= 2;
				len  = *((short *)p); p+= 2;
				for (j=offs; j<offs+len; j++) {
					packets = *(p++);
					x = 0;
					for (k=0; k<packets; k++) {
						skip = *(p++);
						if (skip) x += skip;
						size = *((char *)p); p++;
						if (size>0) {
                            memcpy(thisframe+j*flihead.width+x,p,size);
							p += (int)size;
							x += (int)size;
							}
						else if (size<0) {
                            memset(thisframe+j*flihead.width+x,*p,-size);
							x -= (int)size;
							p++;
							}
						}
					}
				break;
			case 13:	/* FLI_BLACK */
                memset(thisframe,0,flihead.width*flihead.height);
				break;
			case 15:	/* FLI_BRUN */
                for (j=0; j<flihead.height; j++) {
					packets = *(p++);
					x = 0;
					for (k=0; k<packets; k++) {
						size = *(p++);
						if (size>0) {
                            memset(thisframe+j*flihead.width+x,*p,size);
							x += (int)size;
							p++;
							}
						else if (size<0) {
                            memcpy(thisframe+j*flihead.width+x,p,-size);
							x -= (int)size;
							p -= (int)size;
							}
						}
					}
				break;
			case 16:	/* FLI_COPY */
                memcpy(thisframe,p,flihead.width*flihead.height);
                p+= flihead.width*flihead.height;
				break;
            case 18:    /* FLI_MINI */
                p += chunk->size;
                break;
			default:
				return 0x1000+chunk->type;
			}
		p = ((unsigned char *)chunk) + (unsigned int)chunk->size;
		}

    free(flibuf);
	return 0;
}

int play_fli(char *s) {
    GrContext Screen, Mov;
    int x,y,i;

    setupfliread(s);
    x = (GrScreenX() - flihead.width) / 2;
    y = (GrScreenY() - flihead.height) / 2;
    GrCreateContext(flihead.width,flihead.height,NULL,&Mov);
    GrSaveContext(&Screen);
    GrSetContext(&Mov);
    while (!kbhit()) {
        for (i=0; i<flihead.frame_count; i++) {
            if (readframe(Mov.gc_baseaddr[0]) != -1)
                GrBitBlt(&Screen,x,y,&Mov,0,0,flihead.width,flihead.height,GrWRITE);
        }
        lseek(flifile,sizeof(flihead),SEEK_SET);
    }
    getch();
    GrSetContext(&Screen);
    GrDestroyContext(&Mov);
    return 0;
}

void main(int argc, char *argv[]) {
    int Width=320, Height=200, i;

    if (argc>1) {

        /* Parse the command line */
        for (i=2; i < argc; i++) {
            if (strstr(argv[i],"-mode")) {
                Width = atoi(argv[i+1]);
                Height = atoi(argv[i+2]);
            }
        }

        GrSetMode(GR_width_height_color_graphics,Width,Height,256);
        play_fli(argv[1]);
        GrSetMode(GR_default_text);
    }
}
