#include "stdio.h"
#include "stdlib.h"
#include "allegro.h"
#include "mze2dat.h"

/* Define x and y size of Maze -- Must be an ODD number !!! */
#define XSIZE 51 
#define YSIZE 51
#define MAXTREASURE 100
#define FR 250 /* Animation frame rate in miliseconds */

volatile int frme, old_anim;

char score_buf[80];  /* This is where we will sprintf the score before using
                         textout() to write it to scrren */

int midi_count;
int hero_locx, hero_locy, hero_locx2, hero_locy2, h_dead;
int h_canmvn, h_canmvs, h_canmve, h_canmvw;

BITMAP *s; /* Hidden Screen */
BITMAP *s2; /* 2nd Hidden Screen */
DATAFILE *data; /* Allegro DataFile */

char maze[XSIZE][YSIZE];/* Array that holds the maze data */
int objects[XSIZE][YSIZE]; /* Array that holds the object-in-maze data */
int locx, locy, locx2, locy2;

/* Maximum number of enemies in maze at one time */
#define MAX_NASTIES 100

/* Define enemy structure */
volatile struct {
   int x1, y1, x2, y2; /* x1, y1 = Position in maze. x2, y2 = Position on tile */
   int n_type; /* Nasty type. 2=Scorpion, 1=Eyeball */
   int direction; /* Direction Nasty is going in 1,2,3,4 = N,S,E,W */
   int anim_frame; /* Current animation frame of Nasty */
} nasty[MAX_NASTIES];

/* Define hero structure */
volatile struct {
   int x1, y1; /* same as for nasty, but don't need position on tile */
   int direction; /* same as for nasty */
   int h_move; /* is hero moving ? -1=no 1=yes */
   int anim_frame; /* same as for nasty */
   int shot_flag; /* has hero shot bullet ? -1=no, 1=yes */
   int score; /* current score */
} hero;

/* Define bullet structure */
volatile struct {
   int x1, y1, x2, y2; /* same as above , if all = -1 then bullet not active */
   int direction; /* same as above */
   int anim_frame; /* same as above */
   int dist; /* how far bullet has travelled */
} bullet;
  
   
/* Generate Maze - Algorithm from Paul Hart's Web Page */
void makemaze(int x, int y)
	{
	int p = 0, d[4];

	maze[x][y] = ' ';
	if ((x > 1) && (maze[x - 2][y] == 'X')) d[p++] = 1;
	if ((y > 1) && (maze[x][y - 2] == 'X')) d[p++] = 2;
	if ((x < (XSIZE - 2)) && (maze[x + 2][y] == 'X')) d[p++] = 3;
	if ((y < (YSIZE - 2)) && (maze[x][y + 2] == 'X')) d[p++] = 4;
	if (!p) return;
	p = d[rand() % p];
	switch (p)
		{
		case 1: maze[x - 1][y] = ' '; makemaze(x - 2, y); break;
		case 2: maze[x][y - 1] = ' '; makemaze(x, y - 2); break;
		case 3: maze[x + 1][y] = ' '; makemaze(x + 2, y); break;
		case 4: maze[x][y + 1] = ' '; makemaze(x, y + 2); break;
		}
	makemaze(x, y);
	}
	
void gen_maze()
{
	int x,y;
	
	for(x = 0; x < XSIZE; x++)
		for(y = 0; y < YSIZE; y++)
			maze[x][y] = 'X';
	makemaze(1, 1);
}

void gen_init_objs() /* Generate treasure randomly throughout the maze */
{
  int ox, oy, olp1, obj;
  
  for (ox=0; ox < XSIZE; ox++)
  {
   for (oy=0; oy < YSIZE; oy++)
   { 
    objects[ox][oy]=0;
   }
   }
   /* Randomly create treasure in the maze */
  olp1=0;
  do
  {
  ox=rand() / (RAND_MAX/XSIZE+1);
  oy=rand() / (RAND_MAX/YSIZE+1); 
  obj=(rand() / (RAND_MAX/4)) + 7; /* type of treasure */
  
  /* If ox and oy are on a path and location is not the same as hero
     then add treasure type to array objects[][] */
     
  if ((maze[ox][oy]==' ') && ((ox !=1) && (oy !=1)))
  {
    olp1++;
    objects[ox][oy]=obj;
  }
  } while (olp1 < MAXTREASURE);
}      
/* Draw part of the maze onto 1st hidden bitmap */
void plt_hid_map(BITMAP *s)
{
   int x1,y1,x2,y2,x3,y3,tile,obj,nsl; 	
   
   /* X and Y location in the array maze[][] */
   x3=locx; 
   y3=locy;
   
   /* x1 and y1 scan across hidden bitmap */
   for (x1=0; x1 < 14; x1++)
   {
   	for (y1=0; y1 < 14; y1++)
   	{
   		tile=3;	  /* set default tile to 3 - water */
   		x2=(x1*32);    
		y2=290-(y1*32); /* where to draw tile */
		
		/* if we aren't outside of the array maze[][]
		   work out what tile should be. ' ' = tile 0 - stone
		   'X' = tile 1 - hedge */
   		if ((x3+x1 < XSIZE) && (y3+y1 < YSIZE) && 
   		   (x3+x1 > -1) && (y3+y1 > -1))
		   if (maze[x3+x1][y3+y1]==' ') tile=0;
		if ((x3+x1 < XSIZE) && (y3+y1 < YSIZE ) &&
		   (x3+x1 > -1) && (y3+y1 > -1))
		   if (maze[x3+x1][y3+y1]=='X') tile=1;  
		   
		/* plot appropriate tile */
		
		if (tile==0)
		  blit(data[FRAME_02].dat, s, 0,0,x2,y2,32,32);
		if (tile==1)
		  blit(data[FRAME_01].dat, s, 0,0,x2,y2,32,32);
		if (tile==3)
		  blit(data[FRAME_03].dat, s, 0,0,x2,y2,32,32);
                
                /* Any treasure here ? */
                obj=objects[x3+x1][y3+y1];
                if (obj !=0) draw_sprite(s,data[obj].dat,x2+8,y2+8);	
                
                    
	}
   }
  
}	

void init_nasties() /* Randomly create MAX_NASTIES of nasties within the maze */
{
 int lp1, ox, oy;

 lp1=0;
  do
  {
   ox=rand() / (RAND_MAX/XSIZE+1);
   oy=rand() / (RAND_MAX/YSIZE+1);
   
  
  /* If ox and oy are on a path and location is not the same as hero
     then create random nasty type */
     
   if ((maze[ox][oy]==' ') && ((ox !=1) && (oy !=1)))
    {
     nasty[lp1].n_type=(rand() / (RAND_MAX/2)) + 1;     
     nasty[lp1].x1=ox;
     nasty[lp1].y1=oy;
     nasty[lp1].x2=8; /* start nasty in middle of tile */
     nasty[lp1].y2=8;
     nasty[lp1].direction=(rand() / (RAND_MAX/4))+1; /* set random direction */
     
     /* the following code just makes sure that the anim_frame is the correct one
         for the direction the nasty is going in */
     if ((nasty[lp1].direction==1) && (nasty[lp1].n_type==2)) nasty[lp1].anim_frame=31;
	 if ((nasty[lp1].direction==2) && (nasty[lp1].n_type==2)) nasty[lp1].anim_frame=33;
	 if ((nasty[lp1].direction==3) && (nasty[lp1].n_type==2)) nasty[lp1].anim_frame=35;
	 if ((nasty[lp1].direction==4) && (nasty[lp1].n_type==2)) nasty[lp1].anim_frame=37;
	 if ((nasty[lp1].direction==1) && (nasty[lp1].n_type==1)) nasty[lp1].anim_frame=14;
	 if ((nasty[lp1].direction==2) && (nasty[lp1].n_type==1)) nasty[lp1].anim_frame=20;
	 if ((nasty[lp1].direction==4) && (nasty[lp1].n_type==1)) nasty[lp1].anim_frame=16;
	 if ((nasty[lp1].direction==3) && (nasty[lp1].n_type==1)) nasty[lp1].anim_frame=18;
    lp1++;	 
    }
  } while (lp1 < MAX_NASTIES);
}      
  

/* the move_nasties code is only slighty different from the cntrl_screen
   routine - the checks on movement are the same */

void move_nasties()
{
 int nsl,n_canmvn,n_canmvs,n_canmve,n_canmvw;
   
   for (nsl=0; nsl<MAX_NASTIES; nsl++)
    {
     n_canmvn=0;
     n_canmvs=0;
     n_canmve=0;
     n_canmvw=0;
     
     if (nasty[nsl].x1!=-1)
      {
     
        if (maze[nasty[nsl].x1][nasty[nsl].y1+1] == ' ') n_canmvn=1;
	if (maze[nasty[nsl].x1][nasty[nsl].y1-1] == ' ') n_canmvs=1;
	if (maze[nasty[nsl].x1+1][nasty[nsl].y1] == ' ') n_canmve=1;
	if (maze[nasty[nsl].x1-1][nasty[nsl].y1] == ' ') n_canmvw=1;
	
	/* if the tile N, W, E, S is hedge but nasty is not > 16 onto current
	   tile, let him move - this is to take account of the fact that
	   the nasty sprite is 16x16 not a single pixel */
	
	if ((maze[nasty[nsl].x1+1][nasty[nsl].y1] == 'X') && (nasty[nsl].x2 < 14)) n_canmve=1;
	if ((maze[nasty[nsl].x1-1][nasty[nsl].y1] == 'X') && (nasty[nsl].x2 > 4)) n_canmvw=1;
	if ((maze[nasty[nsl].x1][nasty[nsl].y1+1] == 'X') && (nasty[nsl].y2 > 4)) n_canmvn=1;
	if ((maze[nasty[nsl].x1][nasty[nsl].y1-1] == 'X') && (nasty[nsl].y2 < 12)) n_canmvs=1; 
	/* without these two lines, as long as nasty's head is above a hedge his
	   body can walk thru the hedge */
	if ((n_canmve==1) && (maze[nasty[nsl].x1+1][nasty[nsl].y1+1]=='X') && (nasty[nsl].y2 > 16)) n_canmve=0; 
        if ((n_canmvw==1) && (maze[nasty[nsl].x1-1][nasty[nsl].y1+1]=='X') && (nasty[nsl].y2 > 16)) n_canmvw=0;	  
	
	if ((nasty[nsl].direction==1) && (n_canmvn==1)) nasty[nsl].y2-=3;
	if ((nasty[nsl].direction==2) && (n_canmvs==1)) nasty[nsl].y2+=3;
	if ((nasty[nsl].direction==3) && (n_canmve==1)) nasty[nsl].x2+=3;
	if ((nasty[nsl].direction==4) && (n_canmvw==1)) nasty[nsl].x2-=3;
	
	/* check to see if we have moved onto a new tile 
	   and if so, reset our position to beginning of tile. */
	   
	if (nasty[nsl].x2 > 32) {
	   nasty[nsl].x1++;
	   nasty[nsl].x2=0;
	   }
	if (nasty[nsl].x2 < 0) {
	   nasty[nsl].x1--;
	   nasty[nsl].x2=32;
	   }
	
	if (nasty[nsl].y2 < 0) { 
	 nasty[nsl].y1++;
	 nasty[nsl].y2=32;
	 }
	if (nasty[nsl].y2 > 32) {
	  nasty[nsl].y1--;
	  nasty[nsl].y2=0;
	  }

 	/* if nasty can no longer move in the direction it is travelling */
 	   	  
	if (((nasty[nsl].direction==1) && (n_canmvn==0)) || 
	   ((nasty[nsl].direction==2) && (n_canmvs==0)) ||
	   ((nasty[nsl].direction==4) && (n_canmvw==0)) ||
	   ((nasty[nsl].direction==3) && (n_canmve==0))) 
	{
	/* then set new direction for nasty to go in */
	 nasty[nsl].direction=(rand() / (RAND_MAX/4))+1;
	 if ((nasty[nsl].direction==1) && (nasty[nsl].n_type==2)) nasty[nsl].anim_frame=31;
	 if ((nasty[nsl].direction==2) && (nasty[nsl].n_type==2)) nasty[nsl].anim_frame=33;
	 if ((nasty[nsl].direction==3) && (nasty[nsl].n_type==2)) nasty[nsl].anim_frame=35;
	 if ((nasty[nsl].direction==4) && (nasty[nsl].n_type==2)) nasty[nsl].anim_frame=37;
	 if ((nasty[nsl].direction==1) && (nasty[nsl].n_type==1)) nasty[nsl].anim_frame=14;
	 if ((nasty[nsl].direction==2) && (nasty[nsl].n_type==1)) nasty[nsl].anim_frame=20;
	 if ((nasty[nsl].direction==4) && (nasty[nsl].n_type==1)) nasty[nsl].anim_frame=16;
	 if ((nasty[nsl].direction==3) && (nasty[nsl].n_type==1)) nasty[nsl].anim_frame=18;
	} 
    }	 
  }
}	

void create_bullet()
{
 bullet.direction=hero.direction;
 bullet.dist=0;
 bullet.x1=hero_locx;
 bullet.y1=hero_locy;

 /* position bullet in front of the gun */
 
 if (bullet.direction==1)
  {
   bullet.x2=hero_locx2;
   bullet.y2=hero_locy2-10;;
   bullet.anim_frame=4;
  }
  if (bullet.direction==2)
   {
    bullet.x2=hero_locx2;
    bullet.y2=hero_locy2+10;
    bullet.anim_frame=4;
   }
  if (bullet.direction==4)
   {
    bullet.x2=hero_locx2-10;
    bullet.y2=hero_locy2;
    bullet.anim_frame=6;
   }
  if (bullet.direction==3)
   {
    bullet.x2=hero_locx2+10;
    bullet.y2=hero_locy2;
    bullet.anim_frame=6;
   }
   play_sample(data[SHOT].dat, 155, 155, 1000, FALSE); /* bang ! */
}   

void move_bullet()
{
/* if bullet hasn't had 15 'moves' then move bullet */
 if ((bullet.direction==1) && (bullet.dist <15)) bullet.y2-=4;
 if ((bullet.direction==2) && (bullet.dist <15)) bullet.y2+=4;
 if ((bullet.direction==3) && (bullet.dist <15)) bullet.x2+=4;
 if ((bullet.direction==4) && (bullet.dist <15)) bullet.x2-=4;
 if (bullet.y2 < 0)
  {
   bullet.y1++;
   bullet.y2=32;
   }
 if (bullet.y2 > 32)
  {
   bullet.y1--;
   bullet.y2=0;
  }
 if (bullet.x2 > 32)
  {
   bullet.x1++;
   bullet.x2=0;
  }
 if (bullet.x2 < 0)
  {
   bullet.x1--;
   bullet.x2=32;
  }
 bullet.dist++;
 /* otherwise, make bullet explode */
 if (bullet.dist> 15) bullet.anim_frame=11;
 if (bullet.dist> 16) bullet.anim_frame=12;
 if (bullet.dist>17) bullet.anim_frame=13;
 if (bullet.dist>18) hero.shot_flag=0; /* reset, so hero can fire again */
} 



/* Routine to see if hero has walked over some treasure */
void detect_tres_collision()
{
  int colide;
  
  colide=0;
  if (objects[hero_locx][hero_locy] !=0)
  {
  /* the hero must be in the middle of the tile in the direction he is
     walking */
   if ((hero.direction==1) && (hero_locy2 == 10)) colide=1;
   if ((hero.direction==2) && (hero_locy2 == 10)) colide=1;
   if ((hero.direction==3) && (hero_locx2 == 10)) colide=1;
   if ((hero.direction==4) && (hero_locx2 == 10)) colide=1; 
  }
  if (colide==1)
    {
      hero.score+=(objects[hero_locx][hero_locy]*10);
      objects[hero_locx][hero_locy]=0;
      play_sample(data[S_TREAS].dat, 155, 155, 1000, FALSE);
    }
}  

void detect_hero_collision() /* has nasty hit hero ? */
{
  int nsl, colide, xdif, ydif;
  for (nsl=0; nsl < MAX_NASTIES; nsl++)
  {
   xdif=32;
   ydif=32;
   colide=0;
   if ((nasty[nsl].x1==hero_locx) && (nasty[nsl].y1==hero_locy) && (nasty[nsl].x1!=-1))
    {
     if (nasty[nsl].x2 > hero_locx2) xdif=nasty[nsl].x2-hero_locx2;
     if (nasty[nsl].x2 < hero_locx2) xdif=hero_locx2-nasty[nsl].x2;
     if (nasty[nsl].y2 > hero_locy2) ydif=nasty[nsl].y2-hero_locy2;
     if (nasty[nsl].y2 < hero_locy2) ydif=hero_locy2-nasty[nsl].y2;
     if ((xdif < 10) && (ydif < 10)) colide=1;
    }
    if (colide==1)
     {
      play_sample(data[S_DOOMED].dat, 55, 55, 1000, FALSE);
      h_dead=1;
    }
  }
}   

void detect_bullet_collision() /* has bullet hit nasty ? */
{
  int nsl, colide, xdif, ydif;
  for (nsl=0; nsl < MAX_NASTIES; nsl++)
  {
   xdif=32;
   ydif=32;
   colide=0;
   if ((nasty[nsl].x1==bullet.x1) && (nasty[nsl].y1==bullet.y1) && (nasty[nsl].x1!=-1))
    {
     if (nasty[nsl].x2 > bullet.x2) xdif=nasty[nsl].x2-bullet.x2;
     if (nasty[nsl].x2 < bullet.x2) xdif=bullet.x2-nasty[nsl].x2;
     if (nasty[nsl].y2 > bullet.y2) ydif=nasty[nsl].y2-bullet.y2;
     if (nasty[nsl].y2 < bullet.y2) ydif=bullet.y2-nasty[nsl].y2;
     if ((xdif < 10) && (ydif < 10)) colide=1;
    
    if (colide==1)
     {
      play_sample(data[S_GAS].dat, 55, 55, 1000, FALSE);
      nasty[nsl].x1=-1;
      nasty[nsl].y1=-1;
      nasty[nsl].y2=-1;
      nasty[nsl].x2=-1;
      nasty[nsl].n_type=-1;
      nasty[nsl].direction=-1;
      nasty[nsl].anim_frame=-1;
      hero.shot_flag=-1;
      hero.score+=200;
    }
   }
  }
}     
      
/* draw part of hidden screen onto visible screen */
void plt_vis_map(BITMAP *s, BITMAP *s2)
{

 int x1, y1, x2, y2, x3, y3, nsl;
 
   x3=locx;
   y3=locy;
   /* x1 and y1 scan across hidden bitmap */
   for (x1=0; x1 < 14; x1++)
   {
   	for (y1=0; y1 < 14; y1++)
   	{
   		x2=(x1*32);    
		y2=290-(y1*32); /* where to draw nasty */
		
                /* Any Nasties here ? */
                for (nsl=0; nsl<MAX_NASTIES; nsl++)
                {
                 if ((nasty[nsl].x1==x3+x1) && (nasty[nsl].y1==y3+y1) && (nasty[nsl].x1!=-1)) 
                    draw_sprite(s,data[nasty[nsl].anim_frame].dat, x2+nasty[nsl].x2, y2+nasty[nsl].y2);
                }    
                /* Bullet Here ? */
                if (hero.shot_flag!=0)
                 {
                  if ((bullet.x1==x3+x1) && (bullet.y1==y3+y1))
                  {
		   if (bullet.direction==1) draw_sprite(s, data[bullet.anim_frame].dat, x2+bullet.x2, y2+bullet.y2);
	 	   if (bullet.direction==2) draw_sprite_v_flip(s, data[bullet.anim_frame].dat, x2+bullet.x2, y2+bullet.y2);
	 	   if (bullet.direction==4) draw_sprite(s, data[bullet.anim_frame].dat, x2+bullet.x2, y2+bullet.y2);
	 	   if (bullet.direction==3) draw_sprite_h_flip(s, data[bullet.anim_frame].dat, x2+bullet.x2, y2+bullet.y2);
	           move_bullet();
	          }                   
                 }   
	}
    }
	blit(s, s2, locx2+64,locy2+64,0,0,320,200); /* copy 1st hidden screen
	                                               to 2nd hidden screen */
	draw_sprite(s2, data[hero.anim_frame].dat, 160,100);
        sprintf(score_buf, "SCORE=%d", (int) hero.score); 
	textout(s2, data[MAIN_FONT].dat, score_buf, 0,0,3); 
	 move_nasties();
	 vsync(); /* wait for vertical retrace, then copy 2nd hidden screen
	             to visible screen */
	blit(s2, screen, 0,0,0,0,320,200);
	detect_tres_collision(); /* has hero walked over treasure? */
	detect_hero_collision(); /* has nasty hit hero ? */
	if (hero.shot_flag!=0) detect_bullet_collision(); /* has bullet hit nasty ? */
	/* End of Game code to go here */

}	

void cntrl_scrn(BITMAP *s, BITMAP *s2)
{
	/* pretty obvious, depending on which key is pressed
	   alter our position on the current tile and change hero
	   animation frame and direction, or change music*/
	if (key[KEY_S]) stop_midi();
	if (key[KEY_M]) {
	   midi_count++;
	   stop_midi();
	   if (midi_count > 6) midi_count=1;
	   switch (midi_count)
	     {
	     case 1: play_midi(data[MIDI_001].dat, TRUE); break;
	     case 2: play_midi(data[MIDI_002].dat, TRUE); break;
	     case 3: play_midi(data[MIDI_003].dat, TRUE); break;
	     case 4: play_midi(data[MIDI_004].dat, TRUE); break;
	     case 5: play_midi(data[MIDI_005].dat, TRUE); break;
	     case 6: play_midi(data[MIDI_006].dat, TRUE); break;
	     }
	}     
	
	if ((key[KEY_SPACE]) && (hero.shot_flag==0))
	 {
	  hero.shot_flag=1;
	  create_bullet();
	 }
	/* Set things so that the hero is not moving and that the
	   he cannot move in any direction (N,S,W,E) */
	   
	hero.h_move=-1;
	h_canmvn=0;
	h_canmvs=0;
	h_canmvw=0;
	h_canmve=0;
	
	/* if the tile N, S, E, or W is stone then
	   let the hero be able to move in that direction */
	if (maze[hero_locx][hero_locy+1] == ' ') h_canmvn=1;
	if (maze[hero_locx][hero_locy-1] == ' ') h_canmvs=1;
	if (maze[hero_locx+1][hero_locy] == ' ') h_canmve=1;
	if (maze[hero_locx-1][hero_locy] == ' ') h_canmvw=1;
	
	/* if the tile N, W, E, S is hedge but hero is not > 16 onto current
	   tile, let him move - this is to take account of the fact that
	   the hero sprite is 16x16 not a single pixel */
	
	if ((maze[hero_locx+1][hero_locy] == 'X') && (hero_locx2 < 16)) h_canmve=1;
	if ((maze[hero_locx-1][hero_locy] == 'X') && (hero_locx2 > 0)) h_canmvw=1;
	if ((maze[hero_locx][hero_locy+1] == 'X') && (hero_locy2 > 0)) h_canmvn=1;
	if ((maze[hero_locx][hero_locy-1] == 'X') && (hero_locy2 < 14)) h_canmvs=1;
	/* without these two lines, as long as hero's head is above a hedge his
	   body can walk thru the hedge */
	if ((h_canmve==1) && (maze[hero_locx+1][hero_locy+1]=='X') && (hero_locy2 > 16)) h_canmve=0;
        if ((h_canmvw==1) && (maze[hero_locx-1][hero_locy+1]=='X') && (hero_locy2 > 16)) h_canmvw=0;	  
	
	if ((key[KEY_Q]) && (h_canmvn==1)) {
	 locy2-=2;
	 hero_locy2-=2;
	 hero.h_move=1;
	 if (hero.direction!=1) {
	   hero.direction=1;
	   hero.anim_frame=25;
	   }
	 }
	
	if ((key[KEY_A]) && (h_canmvs==1)) {
	 locy2+=2;
	 hero_locy2+=2;
	 hero.h_move=1;
         if (hero.direction!=2) {
           hero.direction=2;
           hero.anim_frame=27;
           }
         }  	
	if ((key[KEY_O]) && (h_canmvw==1)) {
	 locx2-=2;
	 hero_locx2-=2;
	 hero.h_move=1;
	 if (hero.direction!=4) {
	   hero.direction=4;
	   hero.anim_frame=29;
	   }
	 }  
	
	if ((key[KEY_P]) && (h_canmve)) {
	 locx2+=2;
	 hero_locx2+=2;
	 hero.h_move=1;
	 if (hero.direction!=3) {
	   hero.direction=3;
	   hero.anim_frame=23;
	   }
	 }  
	if (hero.h_move==1)  {
	   if (old_anim!=frme) {
           hero.anim_frame++;
           old_anim=frme;
	 if ((hero.anim_frame==27) && (hero.direction==1)) hero.anim_frame=25;
         if ((hero.anim_frame==29) && (hero.direction==2)) hero.anim_frame=27;
	 if ((hero.anim_frame==31) && (hero.direction==4)) hero.anim_frame=29;
	 if ((hero.anim_frame==25) && (hero.direction==3)) hero.anim_frame=23;
   }  	
   } 
	/* check to see if we have moved onto a new tile 
	   and if so, reset our position to beginning of tile. 
	   also check to make sure that we aren't outside the boundries
	   of the map - if we are, reset to inside the maze.  The second
	   check will become redundant after the code to check that hero
	   hasn't hit a wall has been implemented */
	   
	if (hero_locy2 > 32) {
	   hero_locy--;
	   hero_locy2=0;
	   }
	if (locy2 > 32) {
	 locy--;
	 locy2=0;
	}
	if (hero_locy2 < 0) {
	   hero_locy++;
	   hero_locy2=32;
	   }
	if (locy2 < -32) {
	 locy++;
	 locy2=0;
	}
	if (hero_locx2 < 0) { 
	 hero_locx--;
	 hero_locx2=32;
	 }
	if (locx2 < -32) {
	 locx--;
	 locx2=0;
	}
	if (hero_locx2 > 32) {
	  hero_locx++;
	  hero_locx2=0;
	  }
	if (locx2 > 32) {
	 locx++;
	 locx2=0;
	} 
	
	/* second check - now redundant 
	
	if (locx>(XSIZE + 2)) {
	 locx=XSIZE + 2;
	 locx2-=2;
	 }
	if (locx<  -6) {
	 locx=-6;
	 locx2+=2;
	 }
	if (locy>(YSIZE + 2)) {
	 locy=YSIZE + 2;
	 locy2-=2;
	 }
	if (locy< -5) {
	 locy= -5;
	 locy2+=2;
	 } */
	plt_hid_map(s); 
	plt_vis_map(s, s2);
}	

void init_hero()
{
	hero.x1=0;
	hero.y1=0;
	hero.direction=1;
	hero.h_move=-1;
	hero.anim_frame=24;
	hero.shot_flag=0;
	hero.score=0;
	h_dead=0;
}	
/* Use putpixel to show entire maze on screen at once */
void plt_entire_maze(BITMAP *s)
{
 int px1, py1, px2, py2, col, xoff=2, yoff=2;
 clear(screen);
 px2=100;
 py2=((YSIZE+1)*yoff)+45;
 for (px1=0; px1 < XSIZE; px1++)
     {
     for (py1=0; py1 < YSIZE ; py1++)
     {
        col=238;
	if (maze[px1][py1]=='X') col=10; /* actually puts a block of 4 pixels */
	putpixel(screen, px2, py2, col);
	putpixel(screen, px2+1, py2, col);
	putpixel(screen, px2, py2+1, col);
	putpixel(screen, px2+1, py2+1, col);
	py2-=yoff;
     }
     px2+=xoff;
     py2=((YSIZE+1) * yoff)+45;
     } 
     sprintf(score_buf, "Maniac Mazes !!");
     textout(screen, data[GAME_FONT].dat, score_buf,90,0,3);
     sprintf(score_buf, "by Neil Chinnery");
     textout(screen, data[GAME_FONT].dat, score_buf, 86,20,3);
     sprintf(score_buf, "Q = Up");
     textout(screen, data[GAME_FONT].dat, score_buf, 0,60,3);
     sprintf(score_buf, "A = Down");
     textout(screen, data[GAME_FONT].dat, score_buf, 0,80,3);
     sprintf(score_buf, "O = Left");
     textout(screen, data[GAME_FONT].dat, score_buf, 0,100,3);
     sprintf(score_buf, "P = Right");
     textout(screen, data[GAME_FONT].dat, score_buf, 0,120,3);
     sprintf(score_buf, "M = Change");
     textout(screen, data[GAME_FONT].dat, score_buf, 205,60,3);
     sprintf(score_buf, "Music");
     textout(screen, data[GAME_FONT].dat, score_buf, 257,80,3);
     sprintf(score_buf, "S = Stop");
     textout(screen, data[GAME_FONT].dat, score_buf, 205,100,3);
     sprintf(score_buf, "Music");
     textout(screen, data[GAME_FONT].dat, score_buf, 257,120,3);
     sprintf(score_buf, "Space = Fire & Start Game!");
     textout(screen, data[GAME_FONT].dat, score_buf, 35,160,3);
}          
/* This routine is called on an interrupt 
   timer every 'fr' all it does is flip frme between 0 and 1
   it is used to control the animation of the hero - don't know
   if this is the best way to animate a character with only 2
   frames of animation - but it works (sort of!) */
   
void fr_inc()
{
   frme++;
   if (frme>1) frme=0;
}   
END_OF_FUNCTION(fr_inc);

void main()
	{
	
	char buf[80];
	int lp1, lp2;
	/* initialise allegro, and allegro keyboard and timer functions */
	allegro_init();
	install_keyboard();
	install_timer();
	
	set_gfx_mode(GFX_VGA,320,200,0,0); /* set grafix mode */
	
	/* load allegro datafile, containing tiles, sprites, sound fx & music */
	strcpy(get_filename(buf),"MZE2DAT.DAT");
	data=load_datafile(buf);
	
	/* create our large, hidden bitmap and set pallete to
	   same as graphics in the datafile */
	s=create_bitmap(448, 448); /* create hidden screens */
	s2=create_bitmap(320,200);
	set_pallete(data[PALLETE_001].dat);
	do
	{ 
	gen_maze(); /* generate the maze */
        gen_init_objs(); /* Generate treasure in the maze */	
	/* Initialise location  of tile within array maze[][] and
	   location on tile - the x=-6 and y=-3 are a fudge so that
	   the map is always drawn so that it starts with hero in the bottom, 
	   right-hand corner maze [1][1] */
	locx=-6;
	locy=-3;
	hero_locx=1; /* another fudge - have 3rd set of co-ords for hero */
	hero_locy=1;
	hero_locx2=0;
	hero_locy2=0;
	locx2=0;
	locy2=0;
	old_anim=10; /* anything that won't be the same as frme when we get to
	                the first call of cntrl_scrn */
	plt_entire_maze(s);
	install_sound(DIGI_AUTODETECT, MIDI_AUTODETECT, NULL);
        midi_count=3;
        set_volume(255, 130);
        play_midi(data[MIDI_003].dat, TRUE);  
	do
	{
	if (key[KEY_ESC]) exit(0);
	} while (!key[KEY_SPACE]);
	clear_keybuf();
	init_hero();
	init_nasties();
	frme=0;
	/* lock interrupt hero animation routine and lock variables */
	LOCK_VARIABLE(frme);
	LOCK_FUNCTION(fr_inc);
	/* set up interrupt driven hero animation */
	install_int(fr_inc, FR);
	/* Game Loop */
	do 
	{
	cntrl_scrn(s,s2);
	} while ((!key[KEY_ESC]) && (h_dead==0)); /* until someone presses ESC */
	
      } while (!key[KEY_L]);	
	exit(0);
}