#include <jlib.h>
#include "objects.h"

/* this creates 2 objects, one of which rotates 2x as fast as the other */

#define OBJ_DEMO   1 
#define OBJ_DEMO2  2
void handle_demo(obj_data_rec *myself);
void handle_demo2(obj_data_rec *myself);

sprite_system *ssys;

int main(void)
{
 UBYTE *pal;
 list *object_list;
 buffer_rec *offscreen;

 JLIB_DEBUG_OFF;
 JLIB_ENTER("main");

 screen_set_video_mode(); 

 kb_init();

 /* initialise offscreen buffer */
 offscreen = buff_init(SCREEN_WIDTH,SCREEN_HEIGHT);
 buff_draw_string(offscreen,"One of these objects is rotating at 5 Frames Per Second.",0,SCREEN_HEIGHT/2,55);
 buff_draw_string(offscreen,"The other is rotating at 10 Frames Per Second.",0,SCREEN_HEIGHT/2+CHAR_HEIGHT('a'),55);

 ssys=sprite_init(64,64);

 if(sprite_load("demo1.spr",ssys)!=SUCCESS){
    jlib_exit("Couldn't load sprites");
 }
 pal=pal_load("demo1.pal");

 screen_block_set_pal(pal);

 object_list=list_init();

 timing_set_fps(); /* MUST do this before adding objects */

 /* add a demo object (sends itself an init message) */
 objects_add_obj(OBJ_DEMO,handle_demo,100,100,object_list);

 /* add a demo2 object (sends itself an init message) */
 objects_add_obj(OBJ_DEMO2,handle_demo2,100,150,object_list);


 do{
   /* save the area under the sprites */
   buff_save_all_sprites(ssys,offscreen);

   /* draw the sprites */
   buff_draw_all_sprites(ssys,offscreen);

   /* update the screen */
   screen_blit_fs_buffer(offscreen);
   timing_update_fps();

   /* restore the area under the sprites */
   buff_rest_all_sprites(ssys,offscreen);

   /* have each object update itself */
   objects_update_message(object_list,OBJ_MESS_RESET_SPEED);

   /* update the sprites */
   sprite_update_all_anim_and_move(ssys);
 }while(!kb_keydown(KEY_ESC)); 
     
 /* free the offscreen buffer */
 buff_free(offscreen);

 kb_closedown();
 screen_restore_video_mode();

 JLIB_LEAVE;
 return 1;
}


#define DEMO_FIRST_SPRITE 0    /* number of first frame */
#define DEMO_NUM_SPRITES  6    /* number of sprites */
#define DEMO_ANIM_SPEED   10    /* animate at 10 frames per second */

void handle_demo(obj_data_rec *myself)
{
 JLIB_ENTER("handle_demo");

 if(object_get_messages(myself) & OBJ_MESS_INITIALISE_YOURSELF){
    /* perform initialisation */
    object_clear_all_messages(myself);

    myself->sprite = sprite_find_first_free(ssys);
    sprite_set_xy(ssys,myself->sprite,myself->xpos,myself->ypos);
    sprite_set_an_frame(ssys,myself->sprite,DEMO_FIRST_SPRITE);
    sprite_set_move_info(ssys,myself->sprite,0,0,0); 
    sprite_turn_on(ssys,myself->sprite);

    myself->fp_anim = TMG_FP_FRAMES_TO_MOVE(DEMO_ANIM_SPEED);

    myself->fp_an_count = myself->fp_move_x= myself->fp_move_y=0;
    myself->anim_flags=0;
    myself->min_frame=DEMO_FIRST_SPRITE;
    myself->max_frame=DEMO_FIRST_SPRITE+DEMO_NUM_SPRITES;
 }

 if(object_get_messages(myself) & OBJ_MESS_DESTROY_YOURSELF){
    /* free up the sprite and set the object to be deleted */
    sprite_turn_off(ssys,myself->sprite);
    myself->type = OBJ_AWAITING_DELETION;    
    JLIB_LEAVE;
    return;
 }

 /* respond to a change in the frames per second we are acheiving */
 if(object_get_messages(myself) & OBJ_MESS_RESET_SPEED){    
    object_clear_message(myself,OBJ_MESS_RESET_SPEED);
    myself->fp_anim = TMG_FP_FRAMES_TO_MOVE(DEMO_ANIM_SPEED);
 }
 object_move(myself,ssys);
 JLIB_LEAVE;
} 


#define DEMO2_FIRST_SPRITE 0    /* number of first frame */
#define DEMO2_NUM_SPRITES  6    /* number of sprites */
#define DEMO2_ANIM_SPEED   5    /* animate at 5 frames per second */

void handle_demo2(obj_data_rec *myself)
{
 JLIB_ENTER("handle_demo2");

 if(object_get_messages(myself) & OBJ_MESS_INITIALISE_YOURSELF){
    /* perform initialisation */
    object_clear_all_messages(myself);

    myself->sprite = sprite_find_first_free(ssys);
    sprite_set_xy(ssys,myself->sprite,myself->xpos,myself->ypos);
    sprite_set_an_frame(ssys,myself->sprite,DEMO_FIRST_SPRITE);
    sprite_set_move_info(ssys,myself->sprite,0,0,0); 
    sprite_turn_on(ssys,myself->sprite);

    myself->fp_anim = TMG_FP_FRAMES_TO_MOVE(DEMO2_ANIM_SPEED);

    myself->fp_an_count = myself->fp_move_x= myself->fp_move_y=0;
    myself->anim_flags=0;
    myself->min_frame=DEMO_FIRST_SPRITE;
    myself->max_frame=DEMO_FIRST_SPRITE+DEMO_NUM_SPRITES;
 }

 if(object_get_messages(myself) & OBJ_MESS_DESTROY_YOURSELF){
    /* free up the sprite and set the object to be deleted */
    sprite_turn_off(ssys,myself->sprite);
    myself->type = OBJ_AWAITING_DELETION;    
    JLIB_LEAVE;
    return;
 }

 /* respond to a change in the frames per second we are acheiving */
 if(object_get_messages(myself) & OBJ_MESS_RESET_SPEED){    
    object_clear_message(myself,OBJ_MESS_RESET_SPEED);
    myself->fp_anim = TMG_FP_FRAMES_TO_MOVE(DEMO2_ANIM_SPEED);
 }

 object_move(myself,ssys);
 JLIB_LEAVE;
} 
