#include <jlib.h>
#include "objects.h"

/*
** creates an object that responds to input.
*/

#define OBJ_DEMO 1
void handle_demo(obj_data_rec *myself);
#define RANDOM(u) ( rand() % ( ((u) + 1)  ) )

sprite_system *ssys;

int main(void)
{
 UBYTE *pal;
 list *object_list;
 buffer_rec *offscreen;

 JLIB_DEBUG_ON;
 JLIB_ENTER("main");

 screen_set_video_mode(); 

 kb_init();

 /* initialise offscreen buffer */
 offscreen = buff_init(SCREEN_WIDTH,SCREEN_HEIGHT);

 buff_draw_string(offscreen,"This object moves dynamically - use left and right to change speed.",0,SCREEN_HEIGHT/2,55);

 ssys=sprite_init(64,64);

 if(sprite_load("demo1.spr",ssys)!=SUCCESS){
    jlib_exit("Couldn't load sprites");
 }
 pal=pal_load("demo1.pal");

 screen_block_set_pal(pal);

 object_list=list_init();

 timing_set_fps(); /* MUST do this before adding objects */

 /* add a demo object (sends itself an init message) */
 objects_add_obj(OBJ_DEMO,handle_demo,0,0,object_list);

 do{
   /* save the area under the sprites */
   buff_save_all_sprites(ssys,offscreen);

   /* draw the sprites */
   buff_draw_all_sprites(ssys,offscreen);

   /* update the screen */
   screen_blit_fs_buffer(offscreen);
   timing_update_fps();

   /* restore the area under the sprites */
   buff_rest_all_sprites(ssys,offscreen);

   /* have each object update itself */
   objects_update_message(object_list,OBJ_MESS_RESET_SPEED);

   /* update the sprites */
   sprite_update_all_anim_and_move(ssys);
 }while(!kb_keydown(KEY_ESC)); 
     
 /* free the offscreen buffer */
 buff_free(offscreen);

 kb_closedown();
 screen_restore_video_mode();

 JLIB_LEAVE;
 return 1;
}


#define DEMO_FIRST_SPRITE 0    /* number of first frame */
#define DEMO_NUM_SPRITES 6    /* number of sprites */
#define DEMO_ANIM_SPEED 12     /* animate at 12 frames per second */

void handle_demo(obj_data_rec *myself)
{
 JLIB_ENTER("handle_demo");

 if(object_get_messages(myself) & OBJ_MESS_INITIALISE_YOURSELF){
    /* perform initialisation */
    object_clear_all_messages(myself);

    myself->sprite = sprite_find_first_free(ssys);
    sprite_set_xy(ssys,myself->sprite,myself->xpos,myself->ypos);
    sprite_set_an_frame(ssys,myself->sprite,DEMO_FIRST_SPRITE);
    sprite_set_move_info(ssys,myself->sprite,0,0,0); 
    sprite_turn_on(ssys,myself->sprite);

    myself->data[0] = DEMO_ANIM_SPEED;
    myself->fp_anim = TMG_FP_FRAMES_TO_MOVE(myself->data[0]);

    myself->fp_an_count = myself->fp_move_x= myself->fp_move_y=0;
    myself->anim_flags=0;
    myself->min_frame=DEMO_FIRST_SPRITE;
    myself->max_frame=DEMO_FIRST_SPRITE+DEMO_NUM_SPRITES;
 }

 if(object_get_messages(myself) & OBJ_MESS_DESTROY_YOURSELF){
    /* free up the sprite and set the object to be deleted */
    sprite_turn_off(ssys,myself->sprite);
    myself->type = OBJ_AWAITING_DELETION;    
    JLIB_LEAVE;
    return;
 }

 /* respond to a change in the frames per second we are acheiving */
 if(object_get_messages(myself) & OBJ_MESS_RESET_SPEED){    
    object_clear_message(myself,OBJ_MESS_RESET_SPEED);
    myself->fp_anim = TMG_FP_FRAMES_TO_MOVE(myself->data[0]);
 }
 if(kb_keydown(KEY_LEFT)){
    if(myself->data[0]>1){
       myself->data[0]--;
    }
 }

 if(kb_keydown(KEY_RIGHT)){
    myself->data[0]++;
 }

 object_move(myself,ssys);
 JLIB_LEAVE;
} 

