/* Copyright 1996  Jonathan Griffiths. See "readme.jon" for detailed terms. */
#include "objects.h"

unsigned int   global_frame_count;
unsigned int   global_time;
unsigned int   global_fp_fps;

#define NUM_FRAMES 20           /* use 20 frames to determine FPS */

void timing_set_fps(void)
{
 buffer_rec *buff;
 int i,j; 
 double fps;

 JLIB_ENTER("timing_set_fps");

 buff = buff_init(SCREEN_WIDTH,SCREEN_HEIGHT);

 uclock_init();
 j=uclock_read();
 
 for(i=0;i<NUM_FRAMES;i++){
     screen_blit_fs_buffer(buff);
 }  
 
 i=uclock_read();

 buff_free(buff);

 fps = ((float)NUM_FRAMES/(float)(i-j))*(float)UCLOCK_TICKS_PER_SECOND;

 global_fp_fps = (int)(fps*(float)TMG_PRECISION);

 global_frame_count=0;
 global_time = uclock_read();
 JLIB_LEAVE;
}


/* update the fps figure based on the time and frames blitted */
void timing_update_fps(void)
{
 int i;
 float fps;

 JLIB_ENTER("timing_update_fps");

 i=uclock_read()-global_time;
 ++global_frame_count;

 fps = ((float)global_frame_count/(float)(i))*(float)UCLOCK_TICKS_PER_SECOND;
 global_fp_fps = (int)(fps*(float)TMG_PRECISION);

 JLIB_LEAVE;
}


/* Reset the global time and frame blit count */
void timing_reset_fps(void)
{
 JLIB_ENTER("timing_reset_fps");
 global_time=uclock_read();
 global_frame_count=0;
 JLIB_LEAVE;
}

void objects_add_obj(unsigned int type,objfunctptr routine,int x,int y,list *alist)
{
 obj_data_rec *new;
 list_node *temp;

 JLIB_ENTER("objects_add_obj");

 new = (obj_data_rec *)malloc(sizeof(obj_data_rec));

 if(new == NULL){
    jlib_exit("malloc failed");
 }   

 /* copy information into object */
 new->xpos = x;
 new->ypos = y;
 new->type = type;
 new->handling_routine = routine;

 /* add object to list */
 temp = list_get_current(alist);
 list_add_to_start(alist,(void *)new);
 list_set_current(alist,temp);

  /* have object initialise itself */
  object_send_message(new,OBJ_MESS_INITIALISE_YOURSELF);
  new->handling_routine(new);

  JLIB_LEAVE;
}


void objects_update(list *alist)
{
 obj_data_rec *temp;
 list_node *node;

 JLIB_ENTER("objects_update");

 /* loop through objects and update them */ 
 node=alist->head;
 while(node!=NULL){
    temp=(obj_data_rec *)list_get_data_pointer(node);

    if(temp == NULL){
       jlib_exit("NULL entry updating objects");       
     }
       
    if(temp->type == OBJ_AWAITING_DELETION){
       list_set_current(alist,node);
       node=node->next;
       list_remove_current(alist);
       list_set_current(alist,node);
    }
    else{
       temp->dirty=0;
       node=node->next;
       temp->handling_routine(temp);
       if(temp->dirty){
        /* flag dirty in dirty rectangle list */
       }
    }
 }
 list_set_current(alist,alist->head);

 JLIB_LEAVE;
}


void objects_update_message(list *alist,unsigned int mess)
{
 obj_data_rec *temp;
 list_node *node;

 JLIB_ENTER("objects_update_message");

 /* loop through objects and update them */ 
 node=alist->head;
 while(node!=NULL){
    temp=(obj_data_rec *)list_get_data_pointer(node);

    if(temp == NULL){
       jlib_exit("NULL entry updating objects");       
     }
       
    if(temp->type == OBJ_AWAITING_DELETION){
       list_set_current(alist,node);
       node=node->next;
       list_remove_current(alist);
       list_set_current(alist,node);
    }
    else{
       object_send_message(temp,mess);
       temp->dirty=0;
       node=node->next;
       temp->handling_routine(temp);
       if(temp->dirty){
         /* flag dirty in dirty rectangle list */
       }
    }
 }
 list_set_current(alist,alist->head);

 JLIB_LEAVE;
}

/* update an object based on x and y movement, animation, firing and action
 * speed values.
 * NOTE: Assumes sprites are being drawn into a SCREEN_WIDTH*SCREEN_HEIGHT
 * (i.e blit_fs_buffer) buffer for movement wrapping. 
 */
void object_move(obj_data_rec *myself,sprite_system *asys)
{
 int new_frame;

 JLIB_ENTER("object_move");

 /* movement: x and y */
 if(myself->fp_move_x){
    myself->fp_mv_cnt_x += myself->fp_move_x;
    if(TMG_REAL_VAL(myself->fp_mv_cnt_x)){
       myself->dirty=1;
    }

    if(myself->x_dir > 0){
       myself->xpos+=TMG_REAL_VAL(myself->fp_mv_cnt_x);
    }
    else{
       myself->xpos-=TMG_REAL_VAL(myself->fp_mv_cnt_x);
    }

    TMG_CLEAR_VAL(myself->fp_mv_cnt_x);
 }

 if(myself->fp_move_y){
    myself->fp_mv_cnt_y += myself->fp_move_y;

    if(TMG_REAL_VAL(myself->fp_mv_cnt_y)){
       myself->dirty=1;
    }

    if(myself->y_dir > 0){
       myself->ypos+=TMG_REAL_VAL(myself->fp_mv_cnt_y);
    }
    else{
       myself->ypos-=TMG_REAL_VAL(myself->fp_mv_cnt_y);
    }

    TMG_CLEAR_VAL(myself->fp_mv_cnt_y);
 }

 if (myself->move_flags & MV_NO_WRAP){
     if(!(myself->move_flags & MV_NO_FALL_OFF)){
	/* have the object destroy itself if it is offscreen */
	if((myself->xpos > SCREEN_WIDTH) || (myself->xpos < -SPR_MAX_X) ||
		(myself->ypos > SCREEN_HEIGHT) || (myself->ypos < -SPR_MAX_Y)){
                object_send_message(myself,OBJ_MESS_DESTROY_YOURSELF);
	}
     }
 }
 else{
	/* wrap around if called for */
	if(myself->xpos > SCREEN_WIDTH){
	    myself->xpos = -SPR_MAX_X;
            myself->dirty=0;
	}
	else{
	    if(myself->xpos < -SPR_MAX_X){
	       myself->xpos = SCREEN_WIDTH;
               myself->dirty=0;
	    }
	}

        if(myself->ypos > SCREEN_HEIGHT){
	   myself->ypos = -SPR_MAX_Y;
           myself->dirty=0;
        }
	else{
	    if(myself->ypos < -SPR_MAX_Y){
               myself->ypos = SCREEN_HEIGHT;
               myself->dirty=0;
	    }
	}
 } 

 sprite_set_xy(asys,myself->sprite,myself->xpos,myself->ypos);

 /* update my own xand y position so other objects can read it */
 myself->xpos = asys->sprites[myself->sprite]->x;
 myself->ypos = asys->sprites[myself->sprite]->y;

 /* animation */
 if(myself->fp_anim){
    /* move to the next frame */
    myself->fp_an_count += myself->fp_anim;
    new_frame=asys->sprites[myself->sprite]->frame+TMG_REAL_VAL(myself->fp_an_count);
    if(TMG_REAL_VAL(myself->fp_an_count)){
       myself->dirty=1;
    }
    if(myself->anim_flags & AN_NO_WRAP){
       if(new_frame > myself->max_frame){
          object_send_message(myself,OBJ_MESS_DESTROY_YOURSELF);
       }
       else{
          asys->sprites[myself->sprite]->frame=new_frame;
       }
    }
    else{
       while(new_frame >= myself->max_frame){
          new_frame -= (myself->max_frame-myself->min_frame);
       }
       asys->sprites[myself->sprite]->frame=new_frame;
    }
    /* clear the significant part of the fp counter */
    TMG_CLEAR_VAL(myself->fp_an_count);
 }

 /* firing */
 if((myself->fp_fire)&&(myself->can_fire==0)){
    myself->fp_fire_count += myself->fp_fire;
    myself->can_fire=TMG_REAL_VAL(myself->fp_fire_count);
    TMG_CLEAR_VAL(myself->fp_fire_count);
 }

 /* user definable action */
 if((myself->fp_act)&&(myself->can_act==0)){
    myself->fp_act_count += myself->fp_act;
    myself->can_act=TMG_REAL_VAL(myself->fp_act_count);
    TMG_CLEAR_VAL(myself->fp_act_count);
 }

 JLIB_LEAVE;
}
