/* Copyright 1996  Jonathan Griffiths. See "readme.jon" for detailed terms. */
#ifndef __OBJECTS_H__  
#define __OBJECTS_H__

#include <jlib.h>
#include "list.h"

extern unsigned int   global_fp_fps;
extern unsigned int   global_time;
extern unsigned int   global_frame_count;

/* these are the different messages that can be sent to an object */
#define OBJ_MESS_INITIALISE_YOURSELF 0x1
#define OBJ_MESS_DESTROY_YOURSELF    0x2
#define OBJ_MESS_RESET_SPEED         0x4

#define OBJ_AWAITING_DELETION 0

/* movement flags */
#define MV_NO_WRAP      0x1     /* Kill object once off screen */
#define MV_NO_FALL_OFF  0x2     /* Don't destroy me when I'm off screen */

/* animation flags */
#define AN_NO_WRAP      0x1     /* Kill object when animation finished */

/* an objfunctptr is a pointer to a function returning nothing. */
typedef void (*objfunctptr)();

/* object data structure */
typedef struct{
       int xpos;                      /* xpos of object */
       int ypos;                      /* ypos of object */

       unsigned int   type;           /* type of object */
       unsigned char  health;         /* health of object */
       unsigned short sprite;         /* number of sprite used by object */

       objfunctptr  handling_routine; /* function to handle the object */

       unsigned int messages;         /* messages to object */
       int dirty;                     /* have we moved or changed? */

       unsigned int fp_anim;          /* fixed point animation */
       unsigned int anim_flags;      
       unsigned int fp_an_count;      
       int          min_frame;
       int          max_frame;

       unsigned int fp_fire;          /* fixed point firing */
       unsigned int fp_fire_count;      
       unsigned int can_fire;

       unsigned int fp_act;           /* fixed point action */
       unsigned int fp_act_count;      
       unsigned int can_act;

       unsigned int move_flags;       /* what happens when offscreen,  etc */
       signed  char x_dir;            /* positive or negative x movement */
       signed  char y_dir;            /* positive or negative y movement */
       unsigned int fp_move_x;        /* fixed point x movement speed */
       unsigned int fp_mv_cnt_x;      /* counter for x movement speed */
       unsigned int fp_move_y;        /* fixed point y movement speed */
       unsigned int fp_mv_cnt_y;      /* counter for y movement speed */

       int data[10];                  /* user data for handling routine */
}obj_data_rec;


/* add an object to a list */
void objects_add_obj(unsigned int type,objfunctptr routine,int x,int y,list *alist);

/* update all the objects in a list */
void objects_update(list *alist);

/* update all the objects in a list, giving them a certain message */
void objects_update_message(list *alist,unsigned int mess);

/* move an object */
void object_move(obj_data_rec *myself,sprite_system *asys);

/* get the type of an object you have apointer to */
#define object_get_type(ob) ((ob)->type)

/* send a message to an object you have apointer to */
#define object_send_message(ob,mess)  (ob)->messages = (ob)->messages | (mess)

/* clear a message from an object you have a pointer to */
#define object_clear_message(ob,mess)  (ob)->messages = (ob)->messages & ~(mess)

/* clear all messages from an object you have apointer to */
#define object_clear_all_messages(ob)  (ob)->messages = (ob)->messages = 0

/* get the messages of an object you have a pointer to */
#define object_get_messages(ob)  ((ob)->messages)

/* timing stuff */
#define TMG_PRECISION (256*256)  /* 16 bits of precision */

/* get the real value from a fixed point representation */
#define TMG_REAL_VAL(x) ((x)>>16)

/* clear the real value of a fixed point representation */
#define TMG_CLEAR_VAL(x) ((x)=(x)&0x0000ffff)

/* calculate the number of frames to move from a desired frames/sec number */
#define TMG_FP_FRAMES_TO_MOVE(x)  (unsigned int)(((float)((float)((x)<<16)) / (float)global_fp_fps)*TMG_PRECISION ) 

/* calculate the number of frames to move from a desired fp frames/sec number */
#define TMG_FP_FP_FRAMES_TO_MOVE(x) (unsigned int)(((float)(float)(x) / (float)global_fp_fps)*TMG_PRECISION ) 

/* calculate the number of pixels to move from a desired pixels/sec number */
#define TMG_FP_PIXELS_TO_MOVE(x)  (unsigned int)(((float)((float)((x)<<16)) / (float)global_fp_fps)*TMG_PRECISION )

/* calculate the number of pixels to move from a desired fp pixels/sec number */
#define TMG_FP_FP_PIXELS_TO_MOVE(x)  (unsigned int)(((float)(float)(x) / (float)global_fp_fps)*TMG_PRECISION )

/* set up the current frames per second */
void timing_set_fps(void);

/* update the current frames per second */
void timing_update_fps(void);

/* Reset the global time and frame blit count */
void timing_reset_fps(void);

#endif




