
  /* /////////////////////////////////////////////////////////////// */
 /* /// BMP_UTIL.C                                              /// */
/* /////////////////////////////////////////////////////////////// */
	/* --- Elmar Vogt - V1.0 - 1.96 - DJGPP C 2.003 --------- */


#include "bmp_conv.h"

/* === palette conversion utilities ================================ */

/* --- palette_make_pbm(BYTE *, BYTE *) ---------------------------- */
/* converts the second argument, a bmp-palette, to a pbm-palette.
	the source consists of 1024 byte, namely 256 sets of 4 byte each,
	arranged in order [BLUE], [GREEN], [RED], [reserved]. (for what
	reason they included the reserved byte, i have no idea!) maximum
	value for each component is 255.

	pbm-palette format is simply 768 byte, 256 set of 3 each with
	[RED], [GREEN], [BLUE]. maximum component value is 63.
	memory for the target _pbm_palette_ must be preallocated. */

void palette_make_pbm(BYTE *pbm_palette, BYTE *bmp_palette){
	int	i, pbm_index=0, bmp_index=0, n= 256;
		/* n is the number of colors present */

	/* assign indices as necessary: */
	for(i=0, bmp_index=2; i<n; i++, bmp_index+=6){
		pbm_palette[pbm_index++]= bmp_palette[bmp_index--];
		pbm_palette[pbm_index++]= bmp_palette[bmp_index--];
		pbm_palette[pbm_index++]= bmp_palette[bmp_index];
	}

	/* re-calibrate for the lower color resolution of xmode: */
	for(i=0; i<3*256; i++)
		pbm_palette[i]/= 4;
}

/* --- palette_make_bmp(BYTE *, BYTE *) ---------------------------- */
/* the counterpart to _palette_make_pbm()_. takes a pbm-palette as
	second argument, and turns this source into the target, a bmp-
	formatted palette, which is returned as first parameter.
	see also _palette_make_pbm_()_.
	*/

void palette_make_bmp(BYTE *bmp_palette, BYTE *pbm_palette){
	int i, bmp_index=0, pbm_index, n= 256;

	/* cell assignment: */
	for(i=0, pbm_index=2; i<n; i++, pbm_index+=5){
		bmp_palette[bmp_index++]= pbm_palette[pbm_index--];
		bmp_palette[bmp_index++]= pbm_palette[pbm_index--];
		bmp_palette[bmp_index++]= pbm_palette[pbm_index];
		bmp_palette[bmp_index++]= 0;
	}

	/* re-calibration of color resolution: */
	for(i=0; i<4*256; i++)
		bmp_palette[i]*= 4;
}

/* --- make_best_palette_match(BYTE*, BYTE *, BYTE *) --------------- */
/* if you want to load and display a bmp-image without destroying the
	palette you've used before with your xlib, you may like to use this
	function. it maps every color of the _source_bmp_palette_ (in bmp
	format, i.e. 4 byte/color!) to the color of _target_pbm_palette_,
	which yields the smallets squared deviation in all three components.
	please note that this procedure is 'lossy': oftentimes two source
	colors will be mapped to the same target color, and from then on
	will be indistinguishable!
	in general, as one owuld expect, the procedure works quite well
	while you work with images with like palettes, like human faces.
	if you use real landscape sceneries, you will notice that you're
	only using a compromise...
	re-calibration is done in _best_color_match()_, cf. there. */

void make_best_palette_match(BYTE* palette_LUT,
	BYTE *target_pbm_palette, BYTE *source_bmp_palette){

	int	i;			/* loop parameter */

	for(i=0; i<256; i++)
		palette_LUT[i]= best_color_match(target_pbm_palette,
			source_bmp_palette[i*4+2], source_bmp_palette[i*4+1],
			source_bmp_palette[i*4]);
}

/* --- best_color_match(BYTE *, int, int, int) --------------------- */
/* _palette_ must point to an array of 768 BYTEs containing a palette
	in xlib-format. _best_color_match()_ will return the index of the
	color, which most closely resembles the color given by the three
	following arguments, which are considered the red, green and blue
	components, resp., of an arbitrary second color.

	comparison is done on a least-square-difference algorithm.

	note: _palette_ is supposed to contain colors where 63 equals
	maximum component intensity (xlib), while the color to compare the
	palette to uses 255 for maximum intensity (pbm and most other
	formats).
	the maximum mismatch appearing is thus 3*(64^2)= 12288. */

int best_color_match(BYTE *palette, int red, int green, int blue){
	int i, index, best_hit, mismatch, mismatch_old= 12288;
	int red_diff, green_diff, blue_diff;

	/* re-calibrate to pbm-format: */
	red/=4;
	green/=4;
	blue/=4;

	for(i=0, index=0; i<256; i++){
		/* for target palette color, calculate the mismatch: */
		red_diff= (red-palette[index++]);
		green_diff= (green-palette[index++]);
		blue_diff= (blue-palette[index++]);
		mismatch= red_diff*red_diff + green_diff*green_diff
			+ blue_diff*blue_diff;

		/* if the mismatch's less than the old one, update it: */
		if(mismatch<mismatch_old){
			best_hit= i;
			mismatch_old= mismatch;
			/* check- if mismatch is zero, we don't need to look any
				further: */
			if(!mismatch)
				i= 256;
		}
	}

	return best_hit;
}

/* === conversion routines long/int <-> intel data ================= */

/* most data necessary to extract from various file formats are stored
	as two (int) or four (long) consecutive bytes, much like C stores
	its variables. hence, theoretically it would be sufficient to write
		my_value= (int) my_address;
	if _my_address_ is a pointer to an address within an array of bytes,
	holding the information desired.
	but there are two points to consider:
	a) while the number of bytes used for a value in bmp or most other
	formats is fixed, this very number may vary from compiler to
	compiler. if your c-compiler assigns 4 byte to an int, you've lost
	when trying to extract a 2 byte-value, not to mention signs, etc.
	b) bmp format is a genuine microsoft-invention, and thus pretty
	closely clings to the intel convention, where the first byte is
	the least significant, the last byte the most significant. OTOH,
	motorola (and most other producers, BTW) use the opposite notation,
	where the first byte is the most significant. now if you tried to
	read a bmp in intel notation with a program compiled on, say, a
	Mac... you guessed it. */

/* --- read_long_intel(BYTE *)-------------------------------------- */
/* turns four consecutive bytes starting with _address_ to a long var,
	which is returned. intel notation is assumed for the source data. */

long read_long_intel(BYTE *address){
	int i;
	long buffer=0;

	for(i=0; i<4; i++){
		buffer+= address[i];			/* add to buffer */
		buffer= buffer<<8;			/* multiply buffer with 256 = 'shift
												buffer one byte left' */
	}

	return buffer;
}

/* --- read_int_intel(BYTE *) -------------------------------------- */
/* turns the two byte at _address_ and _address_+1 into an int,
	assuming LSB is first. return value is signed or unsigned, depending
	on your compiler implementation. */

int read_int_intel(BYTE *address){
	return (address[0] + 256*address[1]);
}

/* --- write_long_intel(BYTE *, unsigned long) --------------------- */
/* turns _value_ into a four byte long in intel notation and writes
	it to memory starting with _address_.

	_value_'d better be unsigned, to avoid trouble with the sign bit. */

void write_long_intel(BYTE *address, unsigned long value){
	int i;

	for(i=0; i<4; i++){
		address[i]= value&255;		/* values of the 8 least significant
												bits written to _address[i]_ */
		value= value>>8;				/* shift _value_ one byte to the
												right */
	}
}

/* --- write_int_intel(BYTE *, unsigned int) ----------------------- */
/* writes an unsigned int to the BYTEs in _address_ and _address_+1,
	resp.

	_value_'d better be unsigned, to avoid trouble with the sign bit.

	if you have 4 byte ints and try to convert a _value_ > 65535, get
	ready for a surprise. */

void write_int_intel(BYTE *address, unsigned int value){

	address[0]= value&255;
	address[1]= value>>8 ;
}

/* allocates _array_size_ BYTE and initializes them to 0: */
/* (_malloc()_ does not do such an initialization!) */

BYTE* byte_array(unsigned long array_size){
	BYTE* address;
	static size_t sizeof_BYTE= sizeof(BYTE);

	address= (BYTE *) calloc(array_size, sizeof_BYTE);

	if(address==NULL){
		x_set_text_mode();
		fprintf(stderr, "Out of Memory Error.\n");
 		fprintf(stderr, "Could not assign an array of %li BYTES!\n",
 			array_size);
 	}

 	return address;
}

long align_long(unsigned long value){
	int modulo= value&3;

	if(modulo){
		value-= modulo;
		value+= 4;
	}

	return value;
}

