  /* /////////////////////////////////////////////////////////////// */
 /* ///   led.h                                                 /// */
/* /////////////////////////////////////////////////////////////// */
	/* --- Elmar Vogt - V1.0 - 2.96 - DJGPP C++ 2.003 ------- */

/* ================================================================= */
/*	interface file with definitions for the LED class.

	this class has several advantages over usual print-functions, or even
	the print function as provided with xlib:
	-you don't need to have any fonts loaded
	-screen displays and updates need no former screen-blanking, work in
	all graphics modes and remain transparent for the background
	(suitable for diagnostics)
	-they look nifty.

	a 7- and a 16-element led-mode are available through the type
	LED_style. easy-to-use functions for display, appearance and cursor
	control are at hand.

	note, that you can establish several led-instances which remain
	completely independent.
	changing size, slant or display mode of one instance, on the other
	hand, can confuse the cursor control (which is only rudimentary,
	anyway) considerably!

	only those characters that can be found in the first column of the
	corresponding led0/led1-structure in led.cc are displayable.
	*/
/* ================================================================= */

/* === check definitions =========================================== */
#ifndef _LED_CLASS_
#define _LED_CLASS_

/* === include files =============================================== */
// #include "eye.h"

#include <stdio.h>

/* stuff for xlib: */
#include <xlib.h>
	/* currently this is from the file _XLIBDJ25.ZIP_, which is
		version 2.5 for use with djgpp v2.0 */
#include <defines.h>
#include <archive.h>
#include <colors.h>

/* type definitions */
	/* general */
typedef	unsigned int	uINT;
typedef	unsigned long	uLONG;

	/* xlib graphics */
#define	MAX_X				(unsigned) ScrnPhysicalPixelWidth
#define	MAX_Y				(unsigned) ScrnPhysicalHeight
#define	BASE				(unsigned) Page0_Offs


/* === type definitions ============================================ */
typedef enum{SEVEN, SIXTEEN} LED_style;

/* === class definition ============================================ */
class LED{

	public:
			// constructor and destructor:
		LED();
		LED(LED_style, BYTE, BYTE, int, BYTE, BYTE, int, BYTE);
		~LED();

			// parameter settings:
		void set_page(BYTE);
		void set_color(BYTE, BYTE);
		void set_size(BYTE, BYTE);
		void set_slant(int);
		void set_style(LED_style);
		void set_spacing(int);

			// output:
		uINT led(uINT, uINT, char *);
		uINT led(char *);

			// cursor control:
		void bs(uINT);
		void lf();
		void cr();
		void home();

			// user utilities:
		uINT column();
		uINT row();
		BYTE printable(char);

	private:

			// number of led characters displayable:
		const uINT no_led0= 43;
		const uINT no_led1= 97;

			// display settings:
		uINT page;
		uINT fg_color, bg_color;
		uINT size_x, size_y;
		int slant;
		uINT style;
		int spacing;
		int pixel_spacing;

			//cursor positions:
		uINT cursor_x, cursor_y;

			// code lookup-tables:
		BYTE lut[2][256];
			// (these consitute a connection between ascii-code to be
			// displayed, and the corresponding array index)

			// single digit display:
		void digit(uINT, uINT, BYTE);
		void segment(uINT, uINT, uINT, uINT, BYTE, uINT);

		void update_spacing();
};
/* ================================================================= */
#endif		/* _LED_CLASS_ */

